#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Converts a mitmproxy dump file to a swagger schema."""
import argparse
import json
import os
import re
import sys
import traceback
import urllib
from typing import Any, Optional, Sequence, Union

import msgpack
import ruamel.yaml
from mitmproxy.exceptions import FlowReadException

from mitmproxy2swagger import console_util, swagger_util
from mitmproxy2swagger.har_capture_reader import HarCaptureReader, har_archive_heuristic
from mitmproxy2swagger.mitmproxy_capture_reader import (
    MitmproxyCaptureReader,
    mitmproxy_dump_file_huristic,
)


def path_to_regex(path):
    # replace the path template with a regex
    path = re.escape(path)
    path = path.replace(r"\{", "(?P<")
    path = path.replace(r"\}", ">[^/]+)")
    path = path.replace(r"\*", ".*")
    return "^" + path + "$"


def strip_query_string(path):
    # remove the query string from the path
    return path.split("?")[0]


def set_key_if_not_exists(dict, key, value):
    if key not in dict:
        dict[key] = value


def progress_callback(progress):
    console_util.print_progress_bar(progress)


def detect_input_format(file_path):
    har_score = har_archive_heuristic(file_path)
    mitmproxy_score = mitmproxy_dump_file_huristic(file_path)
    if "MITMPROXY2SWAGGER_DEBUG" in os.environ:
        print("har score: " + str(har_score))
        print("mitmproxy score: " + str(mitmproxy_score))
    if har_score > mitmproxy_score:
        return HarCaptureReader(file_path, progress_callback)
    return MitmproxyCaptureReader(file_path, progress_callback)

def str2re(re_str):
    return re.compile("^" + re_str + "$")

def get_abs_path(path):
    base_dir = os.getcwd()
    abs_path = os.path.join(base_dir, path)
    return abs_path


def main(override_args: Optional[Sequence[str]] = None):
    parser = argparse.ArgumentParser(
        description="Converts a mitmproxy dump file or HAR to a swagger schema."
    )
    parser.add_argument(
        "-i",
        "--input",
        help="The input mitmproxy dump file or HAR dump file (from DevTools)",
        required=True,
    )
    parser.add_argument(
        "-o",
        "--output",
        help="The output swagger schema file (yaml). If it exists, new endpoints will be added",
        required=True,
    )
    parser.add_argument("-p", "--api-prefix", help="The api prefix", required=True)
    parser.add_argument(
        "-e",
        "--examples",
        action="store_true",
        help="Include examples in the schema. This might expose sensitive information.",
    )
    parser.add_argument(
        "-hd",
        "--headers",
        action="store_true",
        help="Include headers in the schema. This might expose sensitive information.",
    )
    parser.add_argument(
        "-f",
        "--format",
        choices=["flow", "har"],
        help="Override the input file format auto-detection.",
    )
    parser.add_argument(
        "-r",
        "--param-regex",
        default="[0-9]+",
        help="Regex to match parameters in the API paths. Path segments that match this regex will be turned into parameter placeholders.",
    )
    parser.add_argument(
        "-s",
        "--suppress-params",
        action="store_true",
        help="Do not include API paths that have the original parameter values, only the ones with placeholders.",
    )
    parser.add_argument(
        "-eh",
        "--excluded-headers",
        action="extend",
        nargs="+",
        type=str,
        default=[],
        help="Exclude the headers in schema e.g. -eh 'sec-ch-ua-platform' 'cache-control'",
    )   
    parser.add_argument(
        "-pn",
        "--populate-new",
        action="store_true",
        default=False,
        help="Populate doc for new endpoints in actual requests",
    )
    parser.add_argument(
        "-em",
        "--excluded-methods",
        action="extend",
        nargs="+",
        type=str,
        default=[],
        help="Exclude the request with specific method in schema e.g. -em 'OPTIONS' 'PUT'",
    )  
    parser.add_argument(
        "-w",
        "--overwrite",
        action="store_true",
        help="Overwrite the whole swagger file",
    )
    parser.add_argument(
        "-at",
        "--add-tags",
        action="store_true",
        help="Add swagger tag to the endpoint automatically base on the url",
    )
    parser.add_argument(
        "-tj",
        "--tags-json",
        help='The JSON file mapping URLs to tags, the content can be e.g. {"auth":["/api/auth/.+", "/api/account/.+"]}',
    )
    
    args = parser.parse_args(override_args)

    try:
        args.param_regex = str2re(args.param_regex)
    except re.error as e:
        print(
            f"{console_util.ANSI_RED}Invalid path parameter regex: {e}{console_util.ANSI_RESET}"
        )
        sys.exit(1)

    yaml = ruamel.yaml.YAML()

    capture_reader: Union[MitmproxyCaptureReader, HarCaptureReader]
    if args.format == "flow" or args.format == "mitmproxy":
        capture_reader = MitmproxyCaptureReader(args.input, progress_callback)
    elif args.format == "har":
        capture_reader = HarCaptureReader(args.input, progress_callback)
    else:
        capture_reader = detect_input_format(args.input)

    swagger = None

    # try loading the existing swagger file
    if not args.overwrite:
        try:
            with open(get_abs_path(args.output), "r") as f:
                swagger = yaml.load(f)
        except FileNotFoundError:
            print("No existing swagger file found. Creating new one.")
    if swagger is None:
        swagger = ruamel.yaml.comments.CommentedMap(
            {
                "openapi": "3.0.0",
                "info": {
                    "title": args.input + " Mitmproxy2Swagger",
                    "version": "1.0.0",
                },
            }
        )

    if args.add_tags or args.tags_json:
        swagger["tags"] = []

    tags = None
    if args.tags_json:
        with open(get_abs_path(args.tags_json), "r") as f:
            tags = json.load(f)


    # strip the trailing slash from the api prefix
    args.api_prefix = args.api_prefix.rstrip("/")
    args.excluded_headers = [h.lower() for h in args.excluded_headers]
    args.excluded_methods = [h.lower() for h in args.excluded_methods]

    if "servers" not in swagger or swagger["servers"] is None:
        swagger["servers"] = []

    # add the server if it doesn't exist
    if not any(server["url"] == args.api_prefix for server in swagger["servers"]):
        swagger["servers"].append(
            {"url": args.api_prefix, "description": "The default server"}
        )

    if "paths" not in swagger or swagger["paths"] is None:
        swagger["paths"] = {}

    if "x-path-templates" not in swagger or swagger["x-path-templates"] is None:
        swagger["x-path-templates"] = []

    path_templates = []
    for path in swagger["paths"]:
        path_templates.append(path)

    # also add paths from the the x-path-templates array
    if "x-path-templates" in swagger and swagger["x-path-templates"] is not None:
        for path in swagger["x-path-templates"]:
            path_templates.append(path)

    # new endpoints will be added here so that they can be added as comments in the swagger file
    new_path_templates = []
    path_template_regexes = [re.compile(path_to_regex(path)) for path in path_templates]

    def is_param(param_value):
        return args.param_regex.match(param_value) is not None

    def get_suggested_path(path):
        suggested_path = None
        segments = path.split("/")
        has_param = any(is_param(segment) for segment in segments)
        if has_param:
            # replace digit segments with {id}, {id1}, {id2} etc
            new_segments = []
            param_id = 0
            for segment in segments:
                if is_param(segment):
                    param_name = "id" + str(param_id)
                    if param_id == 0:
                        param_name = "id"
                    new_segments.append("{" + param_name + "}")
                    param_id += 1
                else:
                    new_segments.append(segment)
            suggested_path = "/".join(new_segments)
        return suggested_path

    common_prefix = None
    ref_path = None

    try:
        for req in capture_reader.captured_requests():
            # strip the api prefix from the url
            url = req.get_matching_url(args.api_prefix)

            if url is None:
                continue
            method = req.get_method().lower()
            path = strip_query_string(url).removeprefix(args.api_prefix)
            status = req.get_response_status_code()

            # check if the path matches any of the path templates, and save the index
            path_template_index = None
            for i, path_template_regex in enumerate(path_template_regexes):
                if path_template_regex.match(path):
                    path_template_index = i
                    break
            if path_template_index is None:
                if method not in args.excluded_methods:
                    if path not in new_path_templates:
                        new_path_templates.append(path)
                    if args.populate_new:
                        path_templates.append(get_suggested_path(path) or path)
                        path_template_regexes.append(re.compile(path_to_regex(path)))
                        path_template_index = -1
                    else:
                        continue
            
            if method in args.excluded_methods:
                continue

            path_template_to_set = path_templates[path_template_index]
            set_key_if_not_exists(swagger["paths"], path_template_to_set, {})

            if not ref_path:
                ref_path = path_template_to_set
                common_prefix = path_template_to_set
            else:
                path_prefix = os.path.commonpath([ref_path, path_template_to_set])
                if len(path_prefix)<len(common_prefix):
                    common_prefix = path_prefix

            set_key_if_not_exists(
                swagger["paths"][path_template_to_set],
                method,
                {
                    "tags": [],
                    "summary": swagger_util.path_template_to_endpoint_name(
                        method, path_template_to_set
                    ),
                    "responses": {},
                },
            )

            if args.tags_json:
                for tag, endpoints in tags.items():
                    if any(str2re(ep).match(path_template_to_set) for ep in endpoints):
                        if tag not in swagger["paths"][path_template_to_set][method]["tags"]:
                            swagger["paths"][path_template_to_set][method]["tags"].append(tag)
                            swagger['tags'].append(tag)
                            break

            params = swagger_util.url_to_params(url, path_template_to_set)
            if args.headers:
                headers_request = swagger_util.request_to_headers(
                    req.get_request_headers() if not args.excluded_headers else \
                        {h:v for h,v in req.get_request_headers().items() if h.lower() not in args.excluded_headers}
                )
                params.extend(headers_request)
            if params is not None and len(params) > 0:
                set_key_if_not_exists(
                    swagger["paths"][path_template_to_set][method], "parameters", params
                )

            if method not in ["get", "head"]:
                body = req.get_request_body()
                if body is not None:
                    body_val = None
                    content_type = None
                    # try to parse the body as json
                    try:
                        body_val = json.loads(req.get_request_body())
                        content_type = "application/json"
                    except UnicodeDecodeError:
                        pass
                    except json.decoder.JSONDecodeError:
                        pass

                    # try to parse the body as msgpack, if it's not json
                    if body_val is None:
                        try:
                            body_val = msgpack.loads(req.get_request_body())
                            content_type = "application/msgpack"
                        except Exception:
                            pass

                    if content_type is None:
                        # try to parse the body as form data
                        try:
                            body_val_bytes: Any = dict(
                                urllib.parse.parse_qsl(
                                    body, encoding="utf-8", keep_blank_values=True
                                )
                            )
                            body_val = {}
                            did_find_anything = False
                            for key, value in body_val_bytes.items():
                                did_find_anything = True
                                body_val[key.decode("utf-8")] = value.decode("utf-8")
                            if did_find_anything:
                                content_type = "application/x-www-form-urlencoded"
                            else:
                                body_val = None
                        except UnicodeDecodeError:
                            pass

                    if body_val is not None:
                        content_to_set = {
                            "content": {
                                content_type: {
                                    "schema": swagger_util.value_to_schema(body_val)
                                }
                            }
                        }
                        if args.examples:
                            content_to_set["content"][content_type]["example"] = (
                                swagger_util.limit_example_size(body_val)
                            )
                        set_key_if_not_exists(
                            swagger["paths"][path_template_to_set][method],
                            "requestBody",
                            content_to_set,
                        )

            response_body = req.get_response_body()
            if response_body is not None:
                # try parsing the response as json
                try:
                    response_parsed = json.loads(response_body)
                    response_content_type = "application/json"
                except UnicodeDecodeError:
                    response_parsed = None
                except json.decoder.JSONDecodeError:
                    response_parsed = None

                if response_parsed is None:
                    # try parsing the response as msgpack, if it's not json
                    try:
                        response_parsed = msgpack.loads(response_body)
                        response_content_type = "application/msgpack"
                    except Exception:
                        response_parsed = None

                if response_parsed is not None:
                    resp_data_to_set = {
                        "description": req.get_response_reason(),
                        "content": {
                            response_content_type: {
                                "schema": swagger_util.value_to_schema(response_parsed)
                            }
                        },
                    }
                    if args.examples:
                        resp_data_to_set["content"][response_content_type][
                            "example"
                        ] = swagger_util.limit_example_size(response_parsed)
                    if args.headers:
                        resp_data_to_set["headers"] = swagger_util.response_to_headers(
                            req.get_response_headers()
                        )

                    set_key_if_not_exists(
                        swagger["paths"][path_template_to_set][method]["responses"],
                        str(status),
                        resp_data_to_set,
                    )

            if (
                "responses" in swagger["paths"][path_template_to_set][method]
                and len(swagger["paths"][path_template_to_set][method]["responses"])
                == 0
            ):
                # add a default response if there were no responses detected,
                # this is for compliance with the OpenAPI spec
                content_type = (
                    req.get_response_headers().get("content-type") or "text/plain"
                )

                swagger["paths"][path_template_to_set][method]["responses"]["200"] = {
                    "description": "OK",
                    "content": {},
                }

    except FlowReadException as e:
        print(f"Flow file corrupted: {e}")
        traceback.print_exception(*sys.exc_info())
        print(
            f"{console_util.ANSI_RED}Failed to parse the input file as '{capture_reader.name()}'. "
        )
        if not args.format:
            print(
                f"It might happen that the input format as incorrectly detected. Please try using '--format flow' or '--format har' to specify the input format.{console_util.ANSI_RESET}"
            )
        sys.exit(1)
    except ValueError as e:
        print(f"ValueError: {e}")
        # print stack trace
        traceback.print_exception(*sys.exc_info())
        print(
            f"{console_util.ANSI_RED}Failed to parse the input file as '{capture_reader.name()}'. "
        )
        if not args.format:
            print(
                f"It might happen that the input format as incorrectly detected. Please try using '--format flow' or '--format har' to specify the input format.{console_util.ANSI_RESET}"
            )
        sys.exit(1)

    if args.add_tags and ref_path != common_prefix:
        tags = set()
        tag_index = len(common_prefix.split("/")) if common_prefix!="/" else 1
        for path, p_value in swagger["paths"].items():
            for method, m_value in p_value.items():
                m_value["tags"] = [path.split("/")[tag_index]]
                tags.add(m_value["tags"][0])
        swagger["tags"].extend([{"name": tag} for tag in tags])


    new_path_templates.sort()

    # add suggested path templates
    # basically inspects urls and replaces segments containing only numbers with a parameter
    new_path_templates_with_suggestions = []
    origin_paths = []
    path_prefix = "ignore:" if not args.populate_new else "new:"
    for path in new_path_templates:
        # check if path contains number-only segments
        suggested_path = get_suggested_path(path)
        # prepend the suggested path to the new_path_templates list
        if suggested_path is not None:
            if suggested_path not in new_path_templates_with_suggestions:
                new_path_templates_with_suggestions.append(path_prefix + suggested_path)
            if path not in origin_paths:
                origin_paths.append(path)
        else:
            new_path_templates_with_suggestions.append(path_prefix+ path)

    if not args.suppress_params:
        new_path_templates_with_suggestions.extend([path_prefix+op for op in origin_paths])
    else:
        swagger["x-path-templates"][:] = [
            p for p in swagger["x-path-templates"] 
            if p.split("ignore:")[-1].split("new:")[-1].strip() not in origin_paths
        ]

    # remove the ending comments not to add them twice

    # append the contents of new_path_templates_with_suggestions to swagger['x-path-templates']
    swagger["x-path-templates"].extend(new_path_templates_with_suggestions)

    # remove elements already generated
    swagger["x-path-templates"] = [
        path for path in swagger["x-path-templates"] if path.split("ignore:")[-1].strip() not in swagger["paths"]
    ]

    # remove duplicates while preserving order
    def f7(seq):
        seen = set()
        seen_add = seen.add
        return [x for x in seq if not (x in seen or seen_add(x))]

    swagger["x-path-templates"] = f7(swagger["x-path-templates"])

    swagger["x-path-templates"] = ruamel.yaml.comments.CommentedSeq(
        swagger["x-path-templates"]
    )
    swagger["x-path-templates"].yaml_set_start_comment(
        "Remove the ignore: prefix to generate an endpoint with its URL\n"+
        "Lines that are closer to the top take precedence, the matching is greedy\n"+
        "An endpoint prefixed with `new:` is an endpoint with Doc being populated though it is not in x-path-templates"
    )
    # save the swagger file
    with open(args.output, "w") as f:
        yaml.dump(swagger, f)
    print("Done!")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from openapi_spec_validator import validate_spec

from mitmproxy2swagger.testing_util import mitmproxy2swagger_e2e_test


def test_mitmproxy2swagger_compliance_from_mitmproxy_flow_file():
    data = mitmproxy2swagger_e2e_test(
        "testdata/test_flows",
        "https://httpbin.org/",
        [
            "--format",
            "flow",
        ],
    )
    assert data is not None
    validate_spec(data)


def test_mitmproxy2swagger_compliance_from_mitmproxy_flow_file_with_headers():
    data = mitmproxy2swagger_e2e_test(
        "testdata/test_flows",
        "https://httpbin.org/",
        [
            "--format",
            "flow",
            "--headers",
        ],
    )
    assert data is not None
    validate_spec(data)


def test_mitmproxy2swagger_compliance_from_har_file_with_headers():
    data = mitmproxy2swagger_e2e_test(
        "testdata/sklep.lisek.app.har",
        "https://api2.lisek.app/",
        [
            "--format",
            "har",
            "--headers",
        ],
    )
    assert data is not None
    validate_spec(data)


def test_mitmproxy2swagger_compliance_from_form_data_file_with_headers():
    data = mitmproxy2swagger_e2e_test(
        "testdata/form_data_flows",
        "https://httpbin.org/",
        [
            "--format",
            "flow",
            "--headers",
        ],
    )
    assert data is not None
    validate_spec(data)


def test_mitmproxy2swagger_compliance_from_msgpack_file_with_headers():
    data = mitmproxy2swagger_e2e_test(
        "testdata/msgpack_flows",
        "http://localhost:8082/",
        [
            "--format",
            "flow",
            "--headers",
        ],
    )
    assert data is not None
    validate_spec(data)


def test_mitmproxy2swagger_compliance_from_generic_keys_file_with_headers():
    data = mitmproxy2swagger_e2e_test(
        "testdata/generic_keys_flows",
        "http://localhost:8082/",
        [
            "--format",
            "flow",
            "--headers",
        ],
    )
    assert data is not None
    validate_spec(data)



# -*- coding: utf-8 -*-

import tempfile
from typing import Any, List, Optional

import ruamel.yaml as ruamel

from .mitmproxy2swagger import main


def get_nested_key(obj: Any, path: str) -> Any:
    """Gets a nested key from a dict."""
    keys = path.split(".")
    for key in keys:
        if not isinstance(obj, dict):
            return None
        if key not in obj:
            return None
        obj = obj[key]
    return obj


def mitmproxy2swagger_e2e_test(
    input_file: str, url_prefix: str, extra_args: Optional[List[str]] = None
) -> Any:
    """Runs mitmproxy2swagger on the given input file twice, and returns the detected
    endpoints."""
    yaml_tmp_path = tempfile.mktemp(suffix=".yaml", prefix="sklep.lisek.")
    main(
        [
            "-i",
            input_file,
            "-o",
            yaml_tmp_path,
            "-p",
            url_prefix,
        ]
        + (extra_args or [])
    )
    yaml = ruamel.YAML()

    data = None
    # try to parse the file
    with open(yaml_tmp_path, "r") as f:
        data = yaml.load(f.read())
    assert data is not None
    assert "x-path-templates" in data
    assert "servers" in data
    # remove all of the ignore:prefixes in x-path-templates
    data["x-path-templates"] = [
        x.replace("ignore:", "") for x in data["x-path-templates"]
    ]

    # save the file
    with open(yaml_tmp_path, "w") as f:
        yaml.dump(data, f)

    # run mitmproxy2swagger again
    main(
        [
            "-i",
            input_file,
            "-o",
            yaml_tmp_path,
            "-p",
            url_prefix,
        ]
        + (extra_args or [])
    )

    # load the file again
    with open(yaml_tmp_path, "r") as f:
        data = yaml.load(f.read())
    return data



# -*- coding: utf-8 -*-
import os
import typing
from typing import Iterator
from urllib.parse import urlparse

from mitmproxy import http
from mitmproxy import io as iom
from mitmproxy.exceptions import FlowReadException


def mitmproxy_dump_file_huristic(file_path: str) -> int:
    val = 0
    if "flow" in file_path:
        val += 1
    if "mitmproxy" in file_path:
        val += 1
    # read the first 2048 bytes
    with open(file_path, "rb") as f:
        data = f.read(2048)
        # if file contains non-ascii characters after remove EOL characters
        if (
            data.decode("utf-8", "ignore")
            .replace("\r", "")
            .replace("\n", "")
            .isprintable()
            is False
        ):
            val += 50
        # if first character of the byte array is a digit
        if data[0:1].decode("utf-8", "ignore").isdigit() is True:
            val += 5
        # if it contains the word status_code
        if b"status_code" in data:
            val += 5
        if b"regular" in data:
            val += 10
    return val


class MitmproxyFlowWrapper:
    def __init__(self, flow: http.HTTPFlow):
        self.flow = flow

    def get_url(self) -> str:
        return self.flow.request.url

    def get_matching_url(self, prefix) -> typing.Union[str, None]:
        """Get the requests URL if the prefix matches the URL, None otherwise.

        This takes into account a quirk of mitmproxy where it sometimes
        puts the raw IP address in the URL instead of the hostname. Then
        the hostname is in the Host header.
        """
        if self.flow.request.url.startswith(prefix):
            return self.flow.request.url
        # All the stuff where the real hostname could be
        replacement_hostnames = [
            self.flow.request.headers.get("Host", ""),
            self.flow.request.host_header,
            self.flow.request.host,
        ]
        for replacement_hostname in replacement_hostnames:
            if replacement_hostname is not None and replacement_hostname != "":
                fixed_url = (
                    urlparse(self.flow.request.url)
                    ._replace(netloc=replacement_hostname)
                    .geturl()
                )
                if fixed_url.startswith(prefix):
                    return fixed_url
        return None

    def get_method(self) -> str:
        return self.flow.request.method

    def get_request_headers(self) -> dict[str, typing.List[str]]:
        headers: dict[str, typing.List[str]] = {}
        for k, v in self.flow.request.headers.items(multi=True):
            # create list on key if it does not exist
            headers[k] = headers.get(k, [])
            headers[k].append(v)
        return headers

    def get_request_body(self):
        return self.flow.request.content

    def get_response_status_code(self):
        return self.flow.response.status_code

    def get_response_reason(self):
        return self.flow.response.reason

    def get_response_headers(self):
        headers = {}
        for k, v in self.flow.response.headers.items(multi=True):
            # create list on key if it does not exist
            headers[k] = headers.get(k, [])
            headers[k].append(v)
        return headers

    def get_response_body(self):
        return self.flow.response.content


class MitmproxyCaptureReader:
    def __init__(self, file_path, progress_callback=None):
        self.file_path = file_path
        self.progress_callback = progress_callback

    def captured_requests(self) -> Iterator[MitmproxyFlowWrapper]:
        with open(self.file_path, "rb") as logfile:
            logfile_size = os.path.getsize(self.file_path)
            freader = iom.FlowReader(logfile)
            try:
                for f in freader.stream():
                    if self.progress_callback:
                        self.progress_callback(logfile.tell() / logfile_size)
                    if isinstance(f, http.HTTPFlow):
                        if f.response is None:
                            print(
                                "[warn] flow without response: {}".format(f.request.url)
                            )
                            continue
                        yield MitmproxyFlowWrapper(f)
            except FlowReadException as e:
                print(f"Flow file corrupted: {e}")

    def name(self):
        return "flow"



# import ruamel.yaml
import os

file = "/api/xx/sset/iiii?r=/home/index"
file2 = "/apix/bbb/uuuu/xxx?ss=yy"

print(file.split("/"))
print(file2.split("/"))
print("".split("/"))
print("/".split("/"))
# print(os.path.commonprefix([file, file2]))
print(os.path.commonpath([file, file2]))
print(len(os.path.commonpath([file, file2]).split("/")))

path = "/Users/jiyvn/Documents/oas/m2s/m2src/mitmproxy2swagger/mitmproxy2swagger.py"
base_dir = os.getcwd()
abs_path = os.path.join(base_dir, path)
print(f"abs_path: {abs_path}")


# -*- coding: utf-8 -*-
import os
from base64 import b64decode
from typing import Iterator, Union

import json_stream


# a heuristic to determine if a file is a har archive
def har_archive_heuristic(file_path: str) -> int:
    val = 0
    # if has the har extension
    if file_path.endswith(".har"):
        val += 25
    # read the first 2048 bytes
    with open(file_path, "rb") as f:
        data = f.read(2048)
        # if file contains only ascii characters after remove EOL characters
        if (
            data.decode("utf-8", "ignore")
            .replace("\r", "")
            .replace("\n", "")
            .isprintable()
            is True
        ):
            val += 25
        # sign of a JSON file
        if data[0:1] == b"{":
            val += 23
        # sign of Chrome OR Firefox export
        if b'"WebInspector"' in data or b'"Firefox"' in data:
            val += 15
        if b'"entries"' in data:
            val += 15
        if b'"version"' in data:
            val += 15
    return val


class HarFlowWrapper:
    def __init__(self, flow: dict):
        self.flow = flow

    def get_url(self):
        return self.flow["request"]["url"]

    def get_matching_url(self, prefix) -> Union[str, None]:
        """Get the requests URL if the prefix matches the URL, None otherwise."""
        if self.flow["request"]["url"].startswith(prefix):
            return self.flow["request"]["url"]
        return None

    def get_method(self):
        return self.flow["request"]["method"]

    def get_request_headers(self):
        headers = {}
        for kv in self.flow["request"]["headers"]:
            k = kv["name"]
            v = kv["value"]
            # create list on key if it does not exist
            headers[k] = headers.get(k, [])
            headers[k].append(v)

    def get_request_body(self):
        if (
            "request" in self.flow
            and "postData" in self.flow["request"]
            and "text" in self.flow["request"]["postData"]
        ):
            return self.flow["request"]["postData"]["text"]
        return None

    def get_response_status_code(self):
        return self.flow["response"]["status"]

    def get_response_reason(self):
        return self.flow["response"]["statusText"]

    def get_response_headers(self):
        headers = {}
        for kv in self.flow["response"]["headers"]:
            k = kv["name"]
            v = kv["value"]
            # create list on key if it does not exist
            headers[k] = headers.get(k, [])
            headers[k].append(v)
        return headers

    def get_response_body(self):
        if (
            "response" in self.flow
            and "content" in self.flow["response"]
            and "text" in self.flow["response"]["content"]
        ):
            try:
                if (
                    "encoding" in self.flow["response"]["content"]
                    and self.flow["response"]["content"]["encoding"] == "base64"
                ):
                    return b64decode(self.flow["response"]["content"]["text"]).decode()
            except UnicodeDecodeError:
                return None
            return self.flow["response"]["content"]["text"]
        return None


class HarCaptureReader:
    def __init__(self, file_path: str, progress_callback=None):
        self.file_path = file_path
        self.progress_callback = progress_callback

    def captured_requests(self) -> Iterator[HarFlowWrapper]:
        har_file_size = os.path.getsize(self.file_path)
        with open(self.file_path, "r", encoding="utf-8") as f:
            data = json_stream.load(f)
            for entry in data["log"]["entries"].persistent():
                if self.progress_callback:
                    self.progress_callback(f.tell() / har_file_size)
                yield HarFlowWrapper(entry)

    def name(self):
        return "har"



# -*- coding: utf-8 -*-
import urllib
import uuid
from typing import Any, List

VERBS = [
    "add",
    "create",
    "delete",
    "get",
    "attach",
    "detach",
    "update",
    "push",
    "extendedcreate",
    "activate",
]


# generate a name for the endpoint from the path template
# POST /api/v1/things/{id}/create -> POST create thing by id
def path_template_to_endpoint_name(method, path_template):
    path_template = path_template.strip("/")
    segments = path_template.split("/")
    # remove params to a separate array
    params = []
    for idx, segment in enumerate(segments):
        if segment.startswith("{") and segment.endswith("}"):
            params.append(segment)
            segments[idx] = "{}"
    # remove them from the segments
    segments = [segment for segment in segments if segment != "{}"]
    # reverse the segments
    segments.reverse()
    name_parts = []
    for segment in segments:
        if segment in VERBS:
            # prepend to the name_parts
            name_parts.insert(0, segment.lower())
        else:
            name_parts.insert(0, segment.lower())
            break
    for param in params:
        name_parts.append("by " + param.replace("{", "").replace("}", ""))
        break
    return method.upper() + " " + " ".join(name_parts)


# when given an url and its path template, generates the parameters section of the request
def url_to_params(url, path_template):
    path_template = path_template.strip("/")
    segments = path_template.split("/")
    url_segments = url.split("?")[0].strip("/").split("/")
    params = []
    for idx, segment in enumerate(segments):
        if segment.startswith("{") and segment.endswith("}"):
            params.append(
                {
                    "name": segment.replace("{", "").replace("}", ""),
                    "in": "path",
                    "required": True,
                    "schema": {
                        "type": "number" if url_segments[idx].isdigit() else "string"
                    },
                }
            )
    query_string = urllib.parse.urlparse(url).query
    if query_string:
        query_params = urllib.parse.parse_qs(query_string)
        for key in query_params:
            params.append(
                {
                    "name": key,
                    "in": "query",
                    "required": False,
                    "schema": {
                        "type": "number" if query_params[key][0].isdigit() else "string"
                    },
                }
            )
    return params


def request_to_headers(headers: dict[str, List[Any]], add_example: bool = False):
    """When given an url and its path template, generates the parameters section of the
    request."""
    params = []
    if headers:
        for key in headers:
            h = {
                "name": key,
                "in": "header",
                "required": False,
                "schema": {"type": "number" if headers[key][0].isdigit() else "string"},
            }
            if add_example:
                h["example"] = headers[key][0]
            params.append(h)
    return params


def response_to_headers(headers):
    header = {}
    if headers:
        for key in headers:
            header[key] = {
                "description": headers[key][0],
                "schema": {"type": "number" if headers[key][0].isdigit() else "string"},
            }
    return header


def value_to_schema(value):
    # check if value is a number
    if type(value) is int or type(value) is float:
        return {"type": "number"}
    # check if value is a boolean
    elif isinstance(value, bool):
        return {"type": "boolean"}
    # check if value is a string
    elif isinstance(value, str):
        return {"type": "string"}
    # check if value is a list
    elif isinstance(value, list):
        if len(value) == 0:
            return {"type": "array", "items": {}}

        return {"type": "array", "items": value_to_schema(value[0])}
    # check if value is a dict
    elif isinstance(value, dict):
        all_keys_are_numeric = all(is_numeric_string(key) for key in value)
        all_keys_are_uuid = all(is_uuid(key) for key in value)
        keys_are_generic = all_keys_are_numeric or all_keys_are_uuid

        if keys_are_generic and len(value) > 0:
            return {
                "type": "object",
                "additionalProperties": value_to_schema(list(value.values())[0]),
            }
        return {
            "type": "object",
            "properties": {key: value_to_schema(value[key]) for key in value},
        }
    # if it is none, return null
    elif value is None:
        return {"type": "object", "nullable": True}


def is_uuid(key):
    return isinstance(key, str) and is_valid_uuid(key)


def is_numeric_string(key):
    return isinstance(key, str) and key.isnumeric()


def is_valid_uuid(val):
    try:
        uuid.UUID(str(val))
        return True
    except ValueError:
        return False


MAX_EXAMPLE_ARRAY_ELEMENTS = 10
MAX_EXAMPLE_OBJECT_PROPERTIES = 150


# recursively scan an example value and limit the number of elements and properties
def limit_example_size(example):
    if isinstance(example, list):
        new_list = []
        for element in example:
            if len(new_list) >= MAX_EXAMPLE_ARRAY_ELEMENTS:
                break
            new_list.append(limit_example_size(element))
        return new_list
    elif isinstance(example, dict):
        new_dict = {}
        for key in example:
            if len(new_dict) >= MAX_EXAMPLE_OBJECT_PROPERTIES:
                break
            new_dict[key] = limit_example_size(example[key])
        return new_dict
    else:
        return example



# -*- coding: utf-8 -*-
import sys

ANSI_RGB = "\033[38;2;{};{};{}m"
ANSI_RGB_BG = "\033[48;2;{};{};{}m"
ANSI_RED = "\033[31m"
ANSI_RESET = "\033[0m"

RAINBOW_COLORS = [
    (255, 0, 0),
    (255, 127, 0),
    (255, 255, 0),
    (127, 255, 0),
    (0, 255, 0),
    (0, 255, 127),
    (0, 255, 255),
    (0, 127, 255),
    (0, 0, 255),
    (127, 0, 255),
    (255, 0, 255),
    (255, 0, 127),
]


def rgb_interpolate(start, end, progress):
    return tuple(int(start[i] + (end[i] - start[i]) * progress) for i in range(3))


# take a value from 0 to 1 and return an interpolated color from the rainbow
def rainbow_at_position(progress):
    idx_a = int(progress * float(len(RAINBOW_COLORS) - 1))
    idx_b = idx_a + 1
    return rgb_interpolate(
        RAINBOW_COLORS[idx_a],
        RAINBOW_COLORS[idx_b],
        progress * float(len(RAINBOW_COLORS) - 1) - idx_a,
    )


def print_progress_bar(progress=0.0):
    sys.stdout.write("\r")
    progress_bar_contents = ""
    PROGRESS_LENGTH = 30
    blocks = ["▉", "▊", "▋", "▌", "▍", "▎", "▏"]

    for i in range(PROGRESS_LENGTH):
        interpolated = rainbow_at_position(i / PROGRESS_LENGTH)
        # check if should print a full block
        if i < int(progress * PROGRESS_LENGTH):
            interpolated_2nd_half = rainbow_at_position((i + 0.5) / PROGRESS_LENGTH)
            progress_bar_contents += ANSI_RGB.format(*interpolated)
            progress_bar_contents += ANSI_RGB_BG.format(*interpolated_2nd_half)
            progress_bar_contents += "▌"
        # check if should print a non-full block
        elif i < int((progress * PROGRESS_LENGTH) + 0.5):
            progress_bar_contents += ANSI_RESET
            progress_bar_contents += ANSI_RGB.format(*interpolated)
            progress_bar_contents += blocks[
                int((progress * PROGRESS_LENGTH) + 0.5) - i - 1
            ]
        # otherwise, print a space
        else:
            progress_bar_contents += ANSI_RESET
            progress_bar_contents += " "

    progress_bar_contents += ANSI_RESET
    sys.stdout.write("[{}] {:.1f}%".format(progress_bar_contents, progress * 100))
    sys.stdout.flush()
